import sys
import platform
import psutil
import cpuinfo
import subprocess


class OS:
    name = str()
    full_name = str()
    simple_name = str()
    network_name = str()
    bit = int()

    def update_info():
        if   sys.platform == "win32": OS.name = "Windows"
        elif sys.platform == "darwin": OS.name = "MacOS"
        elif sys.platform.startswith("linux"): OS.name = "Linux"
        elif sys.platform.startswith("freebsd"): OS.name = "FreeBSD"

        OS.full_name     = platform.platform()
        OS.simple_name   = platform.platform(terse=True)
        OS.network_name  = platform.node()

        # NOTE: Debian-64 returns x86_64
        if platform.machine().endswith("64"): OS.bit = 64
        elif platform.machine().endswith("86"): OS.bit = 32


class RAM:
    total        = int()
    total_gb     = int()
    available    = int()
    available_gb = int()
    percent      = float()
    used         = int()
    used_gb      = int()
    free         = int()
    free_gb      = int()

    def update_info():
        vm = psutil.virtual_memory()

        RAM.total        = vm.total
        RAM.total_gb     = int(vm.total / 1073741824)
        RAM.available    = vm.available
        RAM.available_gb = int(vm.available / 1073741824)
        RAM.percent      = vm.percent
        RAM.used         = vm.used
        RAM.used_gb      = int(vm.used / 1073741824)
        RAM.free         = vm.free
        RAM.free_gb      = int(vm.free / 1073741824)


class CPU:
    family =        str()
    brand  =        str()
    hz     =        str()
    hz_advertised = str()
    cores =         int()
    percent =       float()

    def update_info():
        cpi = cpuinfo.get_cpu_info()

        CPU.family        = platform.processor()
        CPU.brand         = cpi["brand_raw"]
        CPU.hz            = cpi["hz_actual_friendly"]
        CPU.hz_advertised = cpi["hz_advertised_friendly"]
        CPU.cores         = psutil.cpu_count()
        CPU.percent       = psutil.cpu_percent()


class Disk:
    device     = str()
    filesystem = str()
    total      = int()
    total_gb   = int()
    percent    = float()
    used       = int()
    used_gb    = int()
    free       = int()
    free_gb    = int()

    def update_info():
        d = psutil.disk_partitions()[0]
        ds = psutil.disk_usage(d.mountpoint)

        Disk.device     = d.device
        Disk.filesystem = d.fstype
        Disk.total      = ds.total
        Disk.total_gb   = int(ds.total / 1073741824)
        Disk.percent    = ds.percent
        Disk.used       = ds.used
        Disk.used_gb    = int(ds.used / 1073741824)
        Disk.free       = ds.free
        Disk.free_gb    = int(ds.free / 1073741824)


def print_specs(update=False):
    if update:
        OS.update_info()
        RAM.update_info()
        CPU.update_info()
        Disk.update_info()

    print("Operating System")
    print(f"  Name       : {OS.name}")
    print(f"  Full name  : {OS.full_name}")
    print(f"  Simple na  : {OS.simple_name}")
    print(f"  Network n  : {OS.network_name}")
    print(f"  Bit        : {OS.bit}")
    print("")
    print("Memory")
    print(f"  Usage      : {RAM.percent}%")
    print(f"  Total      : {RAM.total_gb} GB")
    print(f"  Available  : {RAM.available_gb} GB")
    print(f"  Used       : {RAM.used_gb} GB")
    print(f"  Free       : {RAM.free_gb} GB")
    print("")
    print("Processor")
    print(f"  Usage      : {CPU.percent}%")
    print(f"  Hz         : {CPU.hz}")
    print(f"  Cores      : {CPU.cores}")
    print(f"  Family     : {CPU.family}")
    print(f"  Brand      : {CPU.brand}")
    print("")
    print("Disk")
    print(f"  Usage      : {Disk.percent}")
    print(f"  Filesystem : {Disk.filesystem}")
    print(f"  Total      : {Disk.total_gb} GB")
    print(f"  Used       : {Disk.used_gb} GB")
    print(f"  Free       : {Disk.free_gb} GB")



import pygame
import os
import struct

from heat2d import DISPATCHER
from heat2d.math import Vector2
from heat2d.libs.color import Color
from heat2d.libs.utils import sourcepath


class Window:
    def __init__(self, size, title="Heat2D Game Project", icon=None, clear_color=(0, 0, 0)):
        self.engine = DISPATCHER.engine
        self._width = size[0]
        self._height = size[1]
        self._size = size
        self._title = title
        self._icon = icon

        if icon: self.icon = icon
        else: self.icon = os.path.join(sourcepath, "favicon.png")
        self.size = size
        self.title = self._title

        self.fullscreen = False

        if isinstance(clear_color, Color):
            self.clear_color = clear_color
        else:
            self.clear_color = Color(clear_color)

        self.center = Vector2(self.width / 2, self.height / 2)

        self.clock = pygame.time.Clock()
        self.max_fps = 60
        self.fps = self.max_fps
        self.deltatime = 0

    def __repr__(self):
        return f"<heat2d.Window({self.width}x{self.height}, {self.title})>"

    def take_screenshot(self, filepath, area=None):
        snap = pygame.image.fromstring(self.engine.renderer.ctx.screen.read(), self.size, "RGB")
        if area:
            surf = pygame.Surface((area[2]-area[0], area[3]-area[1]))
            surf.blit(snap, (-area[0], -area[1]))
            pygame.image.save(surf, filepath)
        else:
            pygame.image.save(snap, filepath)

    def toggle_fullscreen(self):
        if self.fullscreen:
            pygame.display.set_mode((self._width, self._height), pygame.DOUBLEBUF|pygame.OPENGL)
            self.surface = pygame.Surface((self._width, self._height))
            self.fullscreen = False

        else:
            pygame.display.set_mode((self._width, self._height), pygame.DOUBLEBUF|pygame.OPENGL|pygame.FULLSCREEN)
            self.surface = pygame.Surface((self._width, self._height))
            self.fullscreen = True

    @property
    def width(self): return self._width

    @width.setter
    def width(self, val):
        self._width = val
        pygame.display.set_mode((self._width, self._height), pygame.DOUBLEBUF|pygame.OPENGL)
        self.surface = pygame.Surface((self._width, self._height))

    @property
    def height(self): return self._height

    @height.setter
    def height(self, val):
        self._height = val
        pygame.display.set_mode((self._width, self._height), pygame.DOUBLEBUF|pygame.OPENGL)
        self.surface = pygame.Surface((self._width, self._height))

    @property
    def size(self): return self._size

    @size.setter
    def size(self, val):
        self._size = val
        self._width = self._size[0]
        self._height = self._size[1]
        pygame.display.set_mode((self._width, self._height), pygame.DOUBLEBUF|pygame.OPENGL)
        self.surface = pygame.Surface((self._width, self._height))

    @property
    def title(self): return self._title

    @title.setter
    def title(self, val):
        self._title = val
        pygame.display.set_caption(self._title)

    @property
    def icon(self): return self._icon

    @icon.setter
    def icon(self, val):
        self._icon = val
        pygame.display.set_icon(pygame.image.load(self._icon))



import time

from heat2d import DISPATCHER


class Timer:
    def __init__(self):
        self.funcs = {"every_tick":[], "every_microsec":[], "every_millisec":[],
                      "every_second":[], "every_minute":[], "every_hour":[]}

        DISPATCHER.engine.timers.append(self)

    def __repr__(self):
        return f"<heat2d.Timer()>"

    def every_tick(self, interval, loop=True):
        def inner_decorator(func):
            self.funcs["every_tick"].append({"func":func, "interval":interval, "loop":loop, "done":False, "tick":0})
            return func
        return inner_decorator

    def every_microsec(self, func, interval, loop=True):
        def inner_decorator(func):
            self.funcs["every_microsec"].append({"func":func, "interval":interval, "loop":loop, "done":False, "last":0, "timef":0})
            return func
        return func

    def every_millisec(self, interval, loop=True):
        def inner_decorator(func):
            self.funcs["every_millisec"].append({"func":func, "interval":interval, "loop":loop, "done":False, "last":0, "timef":0})
            return func
        return inner_decorator

    def every_second(self, func, interval, loop=True):
        def inner_decorator(func):
            self.funcs["every_second"].append({"func":func, "interval":interval, "loop":loop, "done":False, "last":0, "timef":0})
            return func
        return func

    def every_minute(self, func, interval, loop=True):
        def inner_decorator(func):
            self.funcs["every_minute"].append({"func":func, "interval":interval, "loop":loop, "done":False, "last":0, "timef":0})
            return func
        return func


    def every_hour(self, func, interval, loop=True):
        def inner_decorator(func):
            self.funcs["every_hour"].append({"func":func, "interval":interval, "loop":loop, "done":False, "last":0, "timef":0})
            return func
        return func

    def update(self):
        for timef in self.funcs["every_tick"]:
            if not timef["done"]:
                timef["tick"] += 1
                if timef["tick"] > timef["interval"]:
                    timef["tick"] = 0
                    timef["func"]()
                    if not timef["loop"]: timef["done"] = True

        for timef in self.funcs["every_microsec"]:
            if not timef["done"]:
                timef["timef"] = (time.time() - timef["last"]) * 1000000
                if timef["timef"] > timef["interval"]:
                    timef["timef"] = 0
                    timef["last"] = time.time()
                    timef["func"]()
                    if not timef["loop"]: timef["done"] = True

        for timef in self.funcs["every_millisec"]:
            if not timef["done"]:
                timef["timef"] = (time.time() - timef["last"]) * 1000
                if timef["timef"] > timef["interval"]:
                    timef["timef"] = 0
                    timef["last"] = time.time()
                    timef["func"]()
                    if not timef["loop"]: timef["done"] = True

        for timef in self.funcs["every_second"]:
            if not timef["done"]:
                timef["timef"] = time.time() - timef["last"]
                if timef["timef"] > timef["interval"]:
                    timef["timef"] = 0
                    timef["last"] = time.time()
                    timef["func"]()
                    if not timef["loop"]: timef["done"] = True

        for timef in self.funcs["every_minute"]:
            if not timef["done"]:
                timef["timef"] = (time.time() - timef["last"]) / 60
                if timef["timef"] > timef["interval"]:
                    timef["timef"] = 0
                    timef["last"] = time.time()
                    timef["func"]()
                    if not timef["loop"]: timef["done"] = True

        for timef in self.funcs["every_hour"]:
            if not timef["done"]:
                timef["timef"] = (time.time() - timef["last"]) / 3600
                if timef["timef"] > timef["interval"]:
                    timef["timef"] = 0
                    timef["last"] = time.time()
                    timef["func"]()
                    if not timef["loop"]: timef["done"] = True



import pygame


class Context:
    def __init__(self, position, size):
        self.x, self.y = position
        self.width, self.height = size
        self.layers = list()
        self.layers.append(Layer(self))
        self.visible = True
        self.effect = None

    def __repr__(self):
        return f"<heat2d.ui.Context(({self.x},{self.y}), {self.width}x{self.height})>"

    def new_layer(self, num=1):
        for i in range(0, num):
            self.layers.append(Layer(self))


class Layer:
    def __init__(self, context):
        self.context = context
        self.surface = pygame.Surface((self.context.width, self.context.height), pygame.SRCALPHA).convert_alpha()
        self._opacity = 255

    def __repr__(self):
        return f"<heat2d.ui.Layer()>"

    def clear(self, color=None):
        if color: self.surface.fill(color)
        else: self.surface.fill((0, 0, 0, 0))

    @property
    def opacity(self): return self._opacity

    @opacity.setter
    def opacity(self, val):
        self._opacity = val
        self.surface.set_alpha(self._opacity)



import pygame
pygame.font.init()
from heat2d import DISPATCHER, HEAT2D_VERSION, PYGAME_VERSION, SDL_VERSION
from heat2d.ui.context import Context
from heat2d.timer import Timer

fps_display = None
fps_graph = None
font16 = pygame.font.SysFont("Calibri", 16)
font14 = pygame.font.SysFont("Calibri", 14)

def toggle_fps_display():
    global fps_display
    if fps_display.visible:
        fps_display.visible = False
    else:
        fps_display.visible = True

def toggle_debug_display():
    global fps_graph
    if fps_graph.visible:
        fps_graph.visible = False
    else:
        fps_graph.visible = True

def init():
    return
    global fps_display, fps_graph
    engine = DISPATCHER["engine"]

    fps_display = Context((0, 0), (40, 18))
    fps_display.new_layer()
    fps_display.layers[0].clear((0, 0, 0, 255))
    fps_display.layers[0].opacity = 180

    fps_graph = Context((0, engine.window.height-150), (410, 150))
    fps_graph.visible = False
    fps_graph.new_layer(2)
    fps_graph.perf = list()
    fps_graph.layers[0].clear((0, 0, 0, 255))
    fps_graph.layers[0].opacity = 180
    fps_graph.effect = GaussianBlur(radius=4)
    fps_graph.layers[2].surface.blit(font16.render(str(engine.window.max_fps), True, (255, 255, 255)), (257, 5))
    fps_graph.layers[2].surface.blit(font16.render(str(engine.window.max_fps//2), True, (255, 255, 255)), (257, 70))
    fps_graph.layers[2].surface.blit(font16.render("0", True, (255, 255, 255)), (257, 132))
    pygame.draw.line(fps_graph.layers[0].surface, (60, 60, 60), (0, 10), (253, 10))
    pygame.draw.line(fps_graph.layers[0].surface, (60, 60, 60), (0, 75), (253, 75))
    pygame.draw.line(fps_graph.layers[0].surface, (60, 60, 60), (0, 138), (253, 138))
    pygame.draw.line(fps_graph.layers[0].surface, (60, 60, 60), (290, 0), (290, 150))
    fps_graph.layers[2].surface.blit(font16.render("Versions", True, (255, 255, 255)), (295, 5))
    fps_graph.layers[2].surface.blit(font14.render(f"Heat2D  : {HEAT2D_VERSION}", True, (255, 255, 255)), (295, 35))
    fps_graph.layers[2].surface.blit(font14.render(f"SDL         : {SDL_VERSION[0]}.{SDL_VERSION[1]}.{SDL_VERSION[2]}", True, (255, 255, 255)), (295, 55))
    fps_graph.layers[2].surface.blit(font14.render(f"Pygame : {PYGAME_VERSION}", True, (255, 255, 255)), (295, 75))

    engine.add(fps_display)
    engine.add(fps_graph)

    timer = Timer()

    @timer.every_millisec(300)
    def callback():
        fps_display.layers[1].clear()
        fps_display.layers[1].surface.blit(font14.render(str(int(engine.window.fps)), True, (0, 255, 0)), (2, 2))

    @timer.every_millisec(30)
    def callback():
        fps = engine.window.fps
        if fps > engine.window.max_fps: fps = engine.window.max_fps
        fps_graph.perf.append(fps)
        if len(fps_graph.perf)>250: fps_graph.perf.pop(0)

        fps_graph.layers[1].clear()

        for i, p in enumerate(fps_graph.perf):
            p *= (60 / engine.window.max_fps)
            a = int((130 - (p * (130/60))) + 460)

            f = int(510-(p*(510/60)))
            if f == 0: color = (0, 255, 0)
            elif f < 255: color = (255-(255-f), 255, 0)
            else:       color = (255, 255-(f-255), 0)

            pygame.draw.line(fps_graph.layers[1].surface, (color[0], color[1], color[2], 100.0), (i, a-450), (i, 600), 2)
            pygame.draw.line(fps_graph.layers[1].surface, (color[0], color[1], color[2], 255.0), (i, a-450), (i, a-447), 2)



import pygame

from heat2d import DISPATCHER
from heat2d.window import Window
from heat2d.renderer import Renderer
from heat2d.stage import Stage
from heat2d.gameobject import GameObject
from heat2d.libs.keys import key_dictionary, button_dictionary, inv_key_dictionary, inv_button_dictionary
from heat2d.errors import NoStageDeclared
from heat2d.ui.context import Context
from heat2d import ui
from heat2d.timer import Timer
from heat2d.math.vector import Vector2


class Engine:
    def __init__(self):
        DISPATCHER.engine = self
        pygame.init()

        self.window = Window((800, 600))
        self.renderer = Renderer()

        self.timers = list()
        self.sounds = list()

        #working with a dictionary might be better than 3 lists
        self.key_states = {k: [0, 0, 0] for k in key_dictionary}
        self.mouse_states = {b: [0, 0, 0] for b in button_dictionary}

        self.current_stage = None
        self.stages = dict()
        self.gameobjects = list()

        self.__is_running = False
        self._pause = False

        #Initialize modules
        ui.init()

    def __repr__(self):
        return f"<heat2d.Engine({self.window})>"

    def pause(self):
        self._pause = True

    def resume(self):
        self._pause = False

    def add(self, comp, *args, **kwargs):
        if isinstance(comp, Context):
            self.renderer.ui_contexts.append(comp)

            return comp

        else:
            st = comp()

            if isinstance(st, Stage):
                self.stages[st.__class__.__name__] = st
                if len(self.stages) == 1: self.current_stage = st.__class__.__name__
                st.created()

                return st

    def change_stage(self, stage):
        if stage in self.stages: self.current_stage = stage
        else: raise NameError(f"{stage} is not found on stages.")

    def key_pressed(self, key):
        if self.key_states[key][1]: return True
        return False

    def key_released(self, key):
        if self.key_states[key][2]: return True
        return False

    def key_held(self, key):
        if self.key_states[key][0]: return True
        return False

    def mouse_pressed(self, button):
        if self.mouse_states[button][1]: return True
        return False

    def mouse_released(self, button):
        if self.mouse_states[button][2]: return True
        return False

    def mouse_held(self, button):
        if self.mouse_states[button][0]: return True
        return False

    def mouse_wheel_up(self):
        if self.mouse_states["wheelup"][1]: return True
        return False

    def mouse_wheel_down(self):
        if self.mouse_states["wheeldown"][1]: return True
        return False

    def handle_events(self):
        self.mouse = Vector2(pygame.mouse.get_pos())
        for k in self.key_states:
            self.key_states[k][1] = 0
            self.key_states[k][2] = 0

        for b in self.mouse_states:
            self.mouse_states[b][1] = 0
            self.mouse_states[b][2] = 0

        self.events = pygame.event.get()
        for event in self.events:
            if event.type == pygame.QUIT: self.__is_running = False

            elif event.type == pygame.KEYDOWN:
                self.key_states[inv_key_dictionary[event.key]][0] = 1
                self.key_states[inv_key_dictionary[event.key]][1] = 1

            elif event.type == pygame.KEYUP:
                self.key_states[inv_key_dictionary[event.key]][0] = 0
                self.key_states[inv_key_dictionary[event.key]][1] = 0
                self.key_states[inv_key_dictionary[event.key]][2] = 1

            elif event.type == pygame.MOUSEBUTTONDOWN:
                self.mouse_states[inv_button_dictionary[event.button]][0] = 1
                self.mouse_states[inv_button_dictionary[event.button]][1] = 1

            elif event.type == pygame.MOUSEBUTTONUP:
                self.mouse_states[inv_button_dictionary[event.button]][0] = 0
                self.mouse_states[inv_button_dictionary[event.button]][1] = 0
                self.mouse_states[inv_button_dictionary[event.button]][2] = 1

    def stop(self):
        self.__is_running = False

    def run(self):
        if not self.current_stage: raise NoStageDeclared("You have to declare at least one stage to run the engine.")

        self.__is_running = True

        while self.__is_running:
            dt = self.window.clock.tick(self.window.max_fps)
            self.window.deltatime = dt/1000
            self.window.fps = self.window.clock.get_fps()
            self.handle_events()

            self.stages[self.current_stage].update()

            for gameobject in self.stages[self.current_stage].gameobjects: gameobject.update()

            for timer in self.timers: timer.update()

            for sound in self.sounds: sound.update()

            self.renderer.render()

        pygame.quit()
        self.renderer.ctx.release()



import pygame
import time

from heat2d import DISPATCHER


master_volume = 1

class Sound:
    def __init__(self, filename):
        DISPATCHER.engine.sounds.append(self)

        self.volume = 1
        self.right = 1
        self.left = 1
        self.source_sound = pygame.mixer.Sound(filename)
        self.channel = pygame.mixer.find_channel()

        self.dynamic = False
        self.source_object = None
        self.focus_object = None
        self.max_distance = 200
        self.balance_factor = 200
        self.balance_max = 0.73

    def __repr__(self):
        return f"<heat2d.Sound({self.filename}, volume={int(self.volume * 100)}%)>"

    def play(self):
        self.channel.play(self.source_sound)

    def set_source(self, source, focus):
        self.source_object = source
        self.focus_object = focus

    def update(self):
        if self.dynamic and self.source_object and self.focus_object:
            b = (self.source_object.position.x - self.focus_object.position.x) / self.balance_factor
            dist = ((self.source_object.position.x - self.focus_object.position.x)**2 + (self.source_object.position.y - self.focus_object.position.y)**2) / self.max_distance**2
            if dist > 1: dist = 1
            volume = 1 - dist

            if b > 0:
                self.left = self.balance_max - b
                if self.left < 0: self.left = 0
                self.right = 1
            elif b < 0:
                self.left = 1
                self.right = self.balance_max - abs(b)
                if self.right < 0: self.right = 0
            else:
                self.left = 1
                self.right = 1

            self.channel.set_volume(self.left * self.volume * volume * master_volume, self.right * self.volume * volume * master_volume)



#       HEAT2D Game Engine
#      GPL 3.0 © Kadir Aksoy
#https://github.com/kadir014/heat2d

import os

from heat2d.libs import utils

from heat2d.libs import specs
specs.OS.update_info()
if "HEAT2D_UPDATE_SPECS_INFO" not in os.environ: os.environ["HEAT2D_UPDATE_SPECS_INFO"] = "1"
if os.environ["HEAT2D_UPDATE_SPECS_INFO"] == "1":
    print("You can see all by using 'heat2d.specs.print_specs()'\n")
    print("Gathering RAM information...")
    specs.RAM.update_info()
    print("Gathering CPU information...")
    specs.CPU.update_info()
    print("Gathering disk information...")
    specs.Disk.update_info()

os.environ['PYGAME_HIDE_SUPPORT_PROMPT'] = "1"
import pygame

class DISPATCHER: pass

if "HEAT2D_DISABLE_ANSI_COLORS" in os.environ: __USE_ANSI_COLORS = False
else: __USE_ANSI_COLORS = True

from heat2d.libs.terminal import REVERSE, RESET, FG, clear

clear()

from heat2d.version import *

from heat2d.libs.color import Color, Palette
from heat2d.engine import     Engine
from heat2d.stage import      Stage
from heat2d.gameobject import GameObject
from heat2d.sprite import     Sprite
from heat2d.timer import      Timer
from heat2d import            audio
from heat2d import            ui
from heat2d import            gl
from heat2d import            networking
from heat2d.math import       Vector2

if "HEAT2D_HIDE_WELCOME_MESSAGE" not in os.environ:
    if __USE_ANSI_COLORS:
        print(f"\u2588" + "\u2580"*20 + f"\u2588")
        print(f"\u2588{FG.lightred}       HEAT2D       {RESET}\u2588 Version : {HEAT2D_VERSION}")
        print(f"\u2588{FG.orange}    GAME ENGINE     {RESET}\u2588 State   : {HEAT2D_VERSION_STATE.capitalize()}")
        print("\u2588" + "\u2584"*20 + "\u2588\n")
    else:
        print(f"\u2588" + "\u2580"*20 + f"\u2588")
        print(f"\u2588       HEAT2D       \u2588 Version : {HEAT2D_VERSION}")
        print(f"\u2588    GAME ENGINE     \u2588 State   : {HEAT2D_VERSION_STATE.capitalize()}")
        print("\u2588" + "\u2584"*20 + "\u2588\n")

del os, pygame, libs, engine, errors, window, gameobject, renderer, sprite, stage, timer, math, REVERSE, RESET, FG



import moderngl
import pygame
from math import radians

from heat2d import DISPATCHER
from heat2d.version import OPENGL_VERSION, SDL_VERSION
from heat2d.gl import Texture, Shader


class Renderer:
    def __init__(self):
        self.engine = DISPATCHER.engine

        self.viewport = [0, 0, self.engine.window.width, self.engine.window.height]

        self.ctx = moderngl.create_context()
        self.ctx.enable(moderngl.BLEND)

        self.texture = Texture(self.engine.window.surface, ctx=self.ctx)

        self.postprocess = Shader(ctx=self.ctx)

        self.gamefbo = self.ctx.framebuffer((self.texture.texobj,))

    def __repr__(self):
        return f"<heat2d.Renderer(OpenGL {OPENGL_VERSION}, SDL {SDL_VERSION})>"

    def render(self):
        self.gamefbo.use()
        self.gamefbo.clear(*self.engine.window.clear_color.normalise().to_tuple())

        for i, gameobject in enumerate(self.engine.stages[self.engine.current_stage].gameobjects):

            if gameobject.visible:

                gameobject.sprite.texture.use()

                gameobject.sprite.shader.set_uniform("rotation", radians(gameobject.sprite.angle))
                gameobject.sprite.shader.set_uniform("position", gameobject.position.x / self.engine.window.width, gameobject.position.y / self.engine.window.height)
                #gameobject.sprite.shader.set_uniform("size", (gameobject.sprite.source_width-gameobject.sprite.width) / self.engine.window.width, (gameobject.sprite.source_height-gameobject.sprite.height) / self.engine.window.height)
                gameobject.sprite.shader.set_uniform("size",
                        gameobject.sprite.width / self.engine.window.width,
                        gameobject.sprite.height / self.engine.window.height)

                if gameobject.sprite.palette:
                    gameobject.sprite.shader.set_uniform("has_palette", True)
                    gameobject.sprite.shader.set_uniform("palette_length", len(gameobject.sprite.palette))
                    gameobject.sprite.shader.set_uniform("palette", gameobject.sprite._palette_buffer)
                else:
                    gameobject.sprite.shader.set_uniform("has_palette", False)

                gameobject.draw()
                gameobject.sprite.shader.render()

        self.ctx.screen.use()
        self.gamefbo.color_attachments[0].use()
        self.postprocess.render()

        pygame.display.flip()



import moderngl
import pygame

from heat2d import DISPATCHER


filters = {
           "nearest" : moderngl.NEAREST,
           "linear"  : moderngl.LINEAR
          }


class Texture:
    def __init__(self, surface, format="RGBA", ctx=None):
        if ctx: self.ctx = ctx
        else: self.ctx = DISPATCHER.engine.renderer.ctx

        self.surface = surface
        self.format = format
        self.location = 0

        self.texobj = self.ctx.texture(
            self.surface.get_size(),
            len(self.format),
            pygame.image.tostring(self.surface, self.format, True)
        )

        self.texobj.repeat_x = False
        self.texobj.repeat_y = False
        self.filter_min = filters["nearest"]
        self.filter_mag = filters["nearest"]
        self.texobj.filter = (self.filter_min, self.filter_mag)

        self.update()

    def __repr__(self):
        return f"<heat2d.gl.Texture()>"

    def update(self, surface=None, format=None):
        if format == None: format = self.format
        if surface == None: surface = self.surface

        texture_data = pygame.image.tostring(surface, format, True)
        self.texobj.write(texture_data)

    def use(self, location=0):
        self.location = location
        self.texobj.use(self.location)

    def set_repeat(self, x=None, y=None):
        if x != None:
            self.texobj.repeat_x = x

        if y != None:
            self.texobj.repeat_y = y

    def set_filter(self, filter, min=False, mag=False):
        if min and mag:
            self.filter_min = filters[filter]
            self.filter_mag = filters[filter]
            self.texobj.filter = (self.filter_min, self.filter_mag)

        elif min:
            self.filter_min = filters[filter]
            self.texobj.filter = (self.filter_min, self.filter_mag)

        elif mag:
            self.filter_mag = filters[filter]
            self.texobj.filter = (self.filter_min, self.filter_mag)



import os
import pygame
import struct
from math import atan2, degrees

from heat2d import DISPATCHER
from heat2d.libs.utils import sourcepath
from heat2d.gl import Texture, Shader


class Sprite:
    def __init__(self, filepath):
        self.engine = DISPATCHER.engine

        self.gameobject = None
        self.filepath = filepath

        source = pygame.image.load(self.filepath)
        self.source_image = pygame.Surface((source.get_width()+2, source.get_height()+2), pygame.SRCALPHA)
        self.source_image.blit(source, (1, 1))
        self.source_size = self.source_image.get_size()
        self.source_width, self.source_height = self.source_size

        self.width = self.source_width
        self.height = self.source_height

        self.texture = Texture(self.source_image)
        self.shader = Shader(
                        vertex   = os.path.join(sourcepath, "shaders", "sprite.vsh"),
                        fragment = os.path.join(sourcepath, "shaders", "sprite.fsh")
                    )

        self.angle = 0

        self.__palette = None
        self._palette_buffer = None

    def __repr__(self):
        return f"<heat2d.Sprite({self.source_width}x{self.source_height})>"

    def scale(self, scale):
        self.width *= scale
        self.height *= scale

    def rotate(self, angle):
        self.angle += angle

    def rotate_to(self, vector):
        #self.angle = (-degrees(atan2(self.gameobject.position.y - vector.y,  self.gameobject.position.x - vector.x)) - 180) % 360
        self.angle = degrees(atan2(self.gameobject.position.y - vector.y,  self.gameobject.position.x - vector.x))

    @property
    def size(self): return (self.width, self.height)

    @size.setter
    def size(self, val):
        self.width = val[0]
        self.height = val[1]

    @property
    def palette(self): return self.__palette

    @palette.setter
    def palette(self, val):
        self.__palette = val

        colors = list()
        for color in self.__palette:
            colors.append(color.r/255)
            colors.append(color.g/255)
            colors.append(color.b/255)
            colors.append(color.a/255)

        for i in range(256-len(self.__palette)):
            colors.append(0.0)
            colors.append(0.0)
            colors.append(0.0)
            colors.append(1.0)

        self._palette_buffer = struct.pack(f"{256*4}f", *colors)



from heat2d import DISPATCHER
from heat2d.gameobject import GameObject


class Stage:
    def __init__(self):
        self.gameobjects = list()

    def __repr__(self):
        return f"<heat2d.Stage({self.__class__.__name__})>"

    def created(self):
        pass

    def update(self):
        pass

    def add(self, gameobject):
        gobj = gameobject(self)
        self.gameobjects.append(gobj)
        gobj.stage = self
        gobj.created()
        return gobj

    def get(self, gameobject):
        for gobj in self.gameobjects:
            if gobj.__class__.__name__ == gameobject: return gobj



from heat2d import DISPATCHER
from heat2d.math import Vector2


class GameObject:
    def __init__(self, stage):
        self.position = Vector2(0, 0)
        self._sprite = None
        self.stage = stage

        self.visible = True

    def __repr__(self):
        return f"<heat2d.GameObject()>"

    def created(self):
        pass

    def update(self):
        pass

    def draw(self):
        pass

    @property
    def sprite(self): return self._sprite

    @sprite.setter
    def sprite(self, val):
        val.gameobject = self
        self._sprite = val



import pygame
import moderngl


HEAT2D_VERSION = "0.0.4"
HEAT2D_VERSION_TUPLE = (0, 0, 4)
HEAT2D_VERSION_STATE = "alpha"
HEAT2D_LICENSE = "GNU General Public License v3.0"

PYGAME_VERSION = pygame.version.ver
PYGAME_VERSION_TUPLE = pygame.version.vernum
SDL_VERSION = pygame.get_sdl_version()

MODERNGL_VERSION = moderngl.__version__
ctx = moderngl.create_standalone_context(size=(1, 1), require=None)
OPENGL_VERSION = ctx.version_code
ctx.release()


del pygame, moderngl, ctx



from math import sqrt, sin, cos, atan2, radians


class Vector2:
    def __init__(self, a, b=None):
        if isinstance(a, (tuple, list)):
            self.x = a[0]
            self.y = a[1]

        elif isinstance(a, (int, float)):
            if b:
                if isinstance(b, (int, float)):
                    self.x = a
                    self.y = b
                else: raise ValueError("Invalid vector argument")
            else:
                self.x = a
                self.y = 0

        else: raise ValueError("Invalid vector argument")

    def __repr__(self):
        return f"<heat2d.Vector2({self.x}, {self.y})>"

    def __add__(self, other):
        if isinstance(other, Vector2):
            return Vector2(self.x + other.x, self.y + other.y)

        elif isinstance(other, (int, float)):
            return Vector2(self.x + other, self.y + other)

        else:
            raise TypeError(f"unsupported operand type(s) for +: 'Vector2' and '{other.__class__.__name__}'")

    def __sub__(self, other):
        if isinstance(other, Vector2):
            return Vector2(self.x - other.x, self.y - other.y)

        elif isinstance(other, (int, float)):
            return Vector2(self.x - other, self.y - other)

        else:
            raise TypeError(f"unsupported operand type(s) for -: 'Vector2' and '{other.__class__.__name__}'")

    def __mul__(self, other):
        if isinstance(other, Vector2):
            return Vector2(self.x * other.x, self.y * other.y)

        elif isinstance(other, (int, float)):
            return Vector2(self.x * other, self.y * other)

        else:
            raise TypeError(f"unsupported operand type(s) for *: 'Vector2' and '{other.__class__.__name__}'")

    def __truediv__(self, other):
        if isinstance(other, Vector2):
            return Vector2(self.x / other.x, self.y / other.y)

        elif isinstance(other, (int, float)):
            return Vector2(self.x / other, self.y / other)

        else:
            raise TypeError(f"unsupported operand type(s) for /: 'Vector2' and '{other.__class__.__name__}'")

    def __floordiv__(self, other):
        if isinstance(other, Vector2):
            return Vector2(self.x // other.x, self.y // other.y)

        elif isinstance(other, (int, float)):
            return Vector2(self.x // other, self.y // other)

        else:
            raise TypeError(f"unsupported operand type(s) for //: 'Vector2' and '{other.__class__.__name__}'")

    def __iadd__(self, other):
        if isinstance(other, Vector2):
            self = Vector2(self.x + other.x, self.y + other.y)

        elif isinstance(other, (int, float)):
            self = Vector2(self.x + other, self.y + other)

        else:
            raise TypeError(f"unsupported operand type(s) for +: 'Vector2' and '{other.__class__.__name__}'")

        return self

    def __isub__(self, other):
        if isinstance(other, Vector2):
            self = Vector2(self.x - other.x, self.y - other.y)

        elif isinstance(other, (int, float)):
            self = Vector2(self.x - other, self.y - other)

        else:
            raise TypeError(f"unsupported operand type(s) for -: 'Vector2' and '{other.__class__.__name__}'")

        return self

    def __imul__(self, other):
        if isinstance(other, Vector2):
            self = Vector2(self.x * other.x, self.y * other.y)

        elif isinstance(other, (int, float)):
            self = Vector2(self.x * other, self.y * other)

        else:
            raise TypeError(f"unsupported operand type(s) for *: 'Vector2' and '{other.__class__.__name__}'")

        return self

    def __itruediv__(self, other):
        if isinstance(other, Vector2):
            self = Vector2(self.x / other.x, self.y / other.y)

        elif isinstance(other, (int, float)):
            self = Vector2(self.x / other, self.y / other)

        else:
            raise TypeError(f"unsupported operand type(s) for /: 'Vector2' and '{other.__class__.__name__}'")

        return self

    def __ifloordiv__(self, other):
        if isinstance(other, Vector2):
            return Vector2(self.x // other.x, self.y // other.y)

        elif isinstance(other, (int, float)):
            return Vector2(self.x // other, self.y // other)

        else:
            raise TypeError(f"unsupported operand type(s) for //: 'Vector2' and '{other.__class__.__name__}'")

        return self

    def __lt__(self, other):
        return self.length() < other.length()

    def __gt__(self, other):
        return self.length() > other.length()

    def __le__(self, other):
        return self.length() <= other.length()

    def __ge__(self, other):
        return self.length() >= other.length()

    def __eq__(self, other):
        return self.length() == other.length()

    def __ne__(self, other):
        return self.length() != other.length()

    def __neg__(self):
        return self * -1

    def to_tuple(self):
        return (self.x, self.y)

    def length(self):
        return sqrt(self.x**2 + self.y**2)

    def distance_to(self, vector):
        return sqrt((vector.x-self.x)**2 + (vector.y-self.y)**2)

    def angle_to(self, vector):
        return atan2((vector.y-self.y), (vector.x-self.x))

    def set_angle(self, angle):
        self.x *= cos(radians(angle))
        self.y *= sin(radians(angle))

    def get_angle(self):
        return atan2(self.y, self.x)

    def rotate(self, angle):
        self.set_angle(self.get_angle() + angle)

    def rotate_around(self, vector, angle):
        angle = radians(-angle)
        s = sin(angle)
        c = cos(angle)

        self.x -= vector.x
        self.y -= vector.y

        xn = self.x * c - self.y * s
        yn = self.x * s + self.y * c

        self.x = xn + vector.x
        self.y = yn + vector.y

    def dot(self, vector):
        return self.x * vector.x + self.y * vector.y

    def inverse(self):
        return Vector2(1 / self.x, 1 / self.y)

    def normalise(self):
        l = self.length()
        if l == 0: return self
        else: return self / l



import os
import struct

from heat2d import DISPATCHER
from heat2d.libs.utils import sourcepath
from heat2d.math.vector import Vector2


class Shader:
    def __init__(self, vertex=None, fragment=None, ctx=None):
        if ctx: self.ctx = ctx
        else: self.ctx = DISPATCHER.engine.renderer.ctx

        self.texture_coords = [0, 1, 1, 1,
                               0, 0, 1, 0]

        self.world_coords = [-1, -1, 1, -1,
                             -1,  1, 1,  1]

        self.render_ind = [0, 1, 2,
                           1, 2, 3]

        if vertex == None: self.vertex = os.path.join(sourcepath, "shaders", "default.vsh")
        else: self.vertex = vertex

        if fragment == None: self.fragment = os.path.join(sourcepath, "shaders", "default.fsh")
        else: self.fragment = fragment

        self.program = self.ctx.program(
            vertex_shader   = open(self.vertex,   "r").read(),
            fragment_shader = open(self.fragment, "r").read()
        )

        vbo =   self.ctx.buffer(struct.pack('8f', *self.world_coords))
        uvmap = self.ctx.buffer(struct.pack('8f', *self.texture_coords))
        ibo =   self.ctx.buffer(struct.pack('6I', *self.render_ind))

        vao_content = [
            (vbo, '2f', 'vert'),
            (uvmap, '2f', 'in_text'),
        ]

        self.vao = self.ctx.vertex_array(self.program, vao_content, ibo)

    def __repr__(self):
        return f"<heat2d.gl.Shader(vertex={self.vertex}, fragment={self.fragment})>"

    def render(self):
        self.vao.render()

    def set_uniform(self, name, *values):
        uniform = self.program.__getitem__(name)

        if isinstance(values[0], int):
            uniform.write(struct.pack(f"{len(values)}i", *values))

        elif isinstance(values[0], float):
            uniform.write(struct.pack(f"{len(values)}f", *values))

        elif isinstance(values[0], bool):
            uniform.write(struct.pack(f"{len(values)}?", *values))

        elif isinstance(values[0], str):
            uniform.write(struct.pack(f"{len(values)}s", *values))

        elif isinstance(values[0], Vector2):
            uniform.write(struct.pack("2f", values[0].x, values[0].y))

        elif isinstance(values[0], bytes):
            uniform.write(values[0])
